
#include <assert.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include "ealp.h"

/* dbg_out should resolve to a function which accepts a
   single string argument and displays it somewhere where it
   can be seen. */
#ifdef _WIN32
/* In Windows, the function is called OutputDebugString. Rather
   than including windows.h, we hack the prototype inline and link
   with kernel32.lib. */
#ifdef UNICODE
__declspec(dllimport) void __stdcall OutputDebugStringW(const wchar_t*);
#define dbg_out OutputDebugStringW
#else /* UNICODE */
__declspec(dllimport) void __stdcall OutputDebugStringA(const char*);
#define dbg_out OutputDebugStringA
#endif /* UNICODE */

#else
#define dbg_out printf

#endif

/* ABS should return the absolute value of the argument,
   ROUND should round the argument to the nearest integer. 
   ROUNDDOWN should return the biggest integer <= (_x)*/
#define ABS(_x) ((_x) >= 0? (_x) : ((_x)*(-1.0)))
#define ROUND(_x) ((long)((_x)+(0.5)))
#define ROUNDDOWN(_x) ((long)(_x))

/* MAX_FLOAT_VAL is the biggest number that can be represented by
   whatever the type float_t is.
   float_t is a typedef of the floating point type to be used. */
#define MAX_FLOAT_VAL (1.7E308)

/* Allocation algorithm type, holding all information needed for
   the single run. */
struct _alloc_t
{
	/* Index of the last row in the matrix */
	uint_t m;
	/* Index of the last column in the matrix */
	uint_t n;
	/* Index of the last column whose p-value is below the total average */
	uint_t lba;
	/* Column of averages, one per row. */
	float_t* avg;
	/* Row of p-values, one for each column. */
	float_t* p;
	/* The matrix with x-values and sums of each row in
	   the last column, and sums of each column in the last
	   row. Therefore, x[m][n] is the sum of the last column,
	   and the sum of the last row at the same time. These
	   two sums must match, clear, eh? */
	uint_t** x;
};

/* Displays the matrix in it's current state on
   the screen */
void alloc_display(alloc_t* a)
{
	char numt[24];
	char numf[] = "%4u ";
	char numfe[] = "%8.4f ";
	uint_t i, j;

	dbg_out("\n");
	for (i = 0; i <= a->m; i++) {
		for (j = 0; j <= a->n; j++) {
			sprintf(numt, numf, a->x[i][j]);
			dbg_out(numt);
		}
		sprintf(numt, numfe, a->avg[i]);
		dbg_out(numt);
		if (i != a->m) {
			sprintf(numt, numfe, a->avg[i] - a->avg[a->m]);
			dbg_out(numt);
		}
		dbg_out("\n");
	}
	dbg_out("\n");
}

/* allocates memory for the matrix */
unsigned int** matrix_alloc(unsigned int rows, 
	unsigned int cols)
{
	unsigned int** x = (unsigned int**)0;
	unsigned int i;

	x = (unsigned int**)calloc(rows + 1, sizeof(unsigned int*));
	if (x != (unsigned int**)0) {
		for (i = 0; i < rows; i++) {
			x[i] = (unsigned int*)calloc(cols, sizeof(unsigned int*));
			if (x[i] == (unsigned int*)0)
				break;
		}
		if (i < rows) {
			while (i != 0) {
				free(x[i-1]);
				i--;
			}
			free(x);
			x = (unsigned int**)0;
		}
	}
	return x;
}

/* frees memory occupied by the matrix */
void matrix_free(uint_t** x, uint_t m)
{
	uint_t i = m+1;
	while (i != 0) {
		free(x[i-1]);
		i--;
	}
	free(x);
}

/* creates the allocation type and initializes it.
   This must be called first, obviously. */
alloc_t* alloc_create(uint_t num_rows, uint_t num_cols,
	float_t* p, uint_t* row_sums, uint_t* col_sums)
{
	alloc_t* a = NULL;
	uint_t i;

	a = (alloc_t*)malloc(sizeof(alloc_t));
	if (a == NULL)
		return NULL;
	memset(a, 0, sizeof(alloc_t));
	a->m = num_rows;
	a->n = num_cols;

	a->x = matrix_alloc(a->m+1, a->n+1);
	if (a->x == NULL) {
		alloc_destroy(a);
		return NULL;
	}
	a->p = (float_t*)calloc(a->n, sizeof(float_t));
	if (a->p == NULL) {
		alloc_destroy(a);
		return NULL;
	}
	a->avg = (float_t*)calloc(a->m+1, sizeof(float_t));
	if (a->avg == NULL) {
		alloc_destroy(a);
		return NULL;
	}
	
	for (i = 0; i < a->n; i++)
		a->p[i] = p[i];
	for (i = 0; i < a->m; i++)
		a->x[i][a->n] = row_sums[i];
	for (i = 0; i < a->n; i++)
		a->x[a->m][i] = col_sums[i];

	return a;
}

/* Destroys the allocation type and frees memory */
void alloc_destroy(alloc_t* a)
{
	if (a) {
		if (a->avg) 
			free(a->avg);
		if (a->p) 
			free(a->p);
		if (a->x) 
			matrix_free(a->x, a->m);
		free(a);
	}
}

/* Calculates the average of a row and places
   the sum in the last column. */
float_t row_average(alloc_t* a, uint_t idx)
{
	uint_t sumq = 0;
	float_t sumqp = 0.0;
	uint_t i;

	for (i = 0; i < a->n; i++) {
		sumqp += (a->x[idx][i] * a->p[i]);
		sumq += a->x[idx][i];
	}
	a->x[idx][a->n] = sumq;
	return sumq != 0? sumqp / sumq : 0.0;
}

/* determines the lba, which is the index of the last
   column whose p is below the global average */
uint_t calculate_lba(float_t* p, uint_t n, float_t lim)
{
	uint_t i;

	for (i = 0; i <= n; i++) {
		if (p[i] >= lim)
			break;
	}
	return i-1;
}

/* selects the row with minimum distance,
	returns the index of the row */
uint_t sel_lo_row(float_t* avg, uint_t m, uint_t lo)
{
	float_t tmp = MAX_FLOAT_VAL;
	uint_t ret = ROW_INVALID;
	uint_t i;

	if (lo == ROW_INVALID) {
		for (i = 0; i < m; i++) {
			if (avg[i] <= avg[m] && avg[i] < tmp) {
				tmp = avg[i];
				ret = i;
			}
		}
	} else {
		for (i = lo + 1; i < m; i++) {
			if (avg[i] <= avg[m] && avg[i] >= avg[lo] && avg[i] < tmp) {
				tmp = avg[i];
				ret = i;
			}
		}
	}

	return ret;
}

/* selects the row with maximum distance,
	returns the index of the row */
uint_t sel_hi_row(float_t* avg, uint_t m, uint_t lo)
{
	float_t tmp = 0.0;
	uint_t ret = ROW_INVALID;
	uint_t i;
	
	if (lo == ROW_INVALID) {
		for (i = 0; i < m; i++) {
			if (avg[i] > avg[m] && avg[i] > tmp) {
				tmp = avg[i];
				ret = i;
			}
		}
	} else {
		for (i = lo + 1; i < m; i++) {
			if (avg[i] > avg[m] && avg[i] >= avg[lo] && (avg[i] > tmp)) {
				tmp = avg[i];
				ret = i;
			}
		}
	}

	return ret;
}

/* blah blah */
uint_t sum_row(uint_t** mtx, uint_t idx, uint_t n)
{
	uint_t sum = 0;
	uint_t j;

	for (j = 0; j < n; j++)
		sum += mtx[idx][j];
	return sum;
}

/* blah blah */
uint_t sum_col(uint_t** mtx, uint_t idx, uint_t m)
{
	uint_t sum = 0;
	uint_t i;

	for (i = 0; i < m; i++)
		sum += mtx[i][idx];
	return sum;
}

/* blah blah */
uint_t find_bad_row(uint_t** mtx, uint_t m, uint_t n)
{
	uint_t ret = ROW_INVALID;
	uint_t sum = 0;
	uint_t i;

	for (i = 0; i < m; i++) {
		sum = sum_row(mtx, i, n);
		if (sum != mtx[i][n]) {
			ret = i;
			break;
		}
	}
	return ret;
}

/* blah blah */
uint_t find_bad_col(uint_t** mtx, uint_t m, uint_t n)
{
	uint_t ret = COL_INVALID;
	uint_t sum = 0;
	uint_t j;

	for (j = 0; j < n; j++) {
		sum = sum_col(mtx, j, m);
		if (sum != mtx[m][j]) {
			ret = j;
			break;
		}
	}

	return ret;
}

error_t stage_init(alloc_t* a)
{
	error_t err = ERR_OK;

	/* total average & lba */
	a->avg[a->m] = row_average(a, a->m);
	a->lba = calculate_lba(a->p, a->n, a->avg[a->m]);

	return err;
}

/* blah blah */
error_t stage_spread(alloc_t* a)
{
	error_t err = ERR_OK;
	float_t* k;
	uint_t* sumrow;
	uint_t i, j;
	unsigned char done;

	k = (float_t*)calloc(a->m, sizeof(float_t));
	if (k != NULL) {
		sumrow = (uint_t*)calloc(a->n, sizeof(uint_t));
		if (sumrow != NULL) {
			/* Copy the sum row */
			for (j = 0; j < a->n; j++)
				sumrow[j] = a->x[a->m][j];
			/* calculate spread factors */
			for (i = 0; i < a->m; i++) {
				/* If mtx[m][n] is zero, the program will die here.
					It should never be zero. */
				k[i] = (float_t)a->x[i][a->n] / (float_t)a->x[a->m][a->n];
			}
			/* spread */
			for (i = 0; i < a->m; i++) {
				for (j = 0; j < a->n; j++) {
					a->x[i][j] = ROUNDDOWN(k[i] * (float_t)a->x[a->m][j]);
					sumrow[j] -= a->x[i][j];
					assert(sumrow[j] >= 0);
				}
			}
			/* adjust */
			done = 0;
			while (!done) {
				done = 1;
				for (j = 0; j < a->n; j++) {
					if (sumrow[j] != 0) {
						i = find_bad_row(a->x, a->m, a->n);
						assert(i != ROW_INVALID);
						a->x[i][j] += 1;
						sumrow[j] -= 1;
						done = 0; break;
					}
				}
			}
			free(sumrow);
		}
		free(k);
	}

	for (i = 0; i < a->m; i++)
		a->avg[i] = row_average(a, i);

	return err;
}

/* blah blah */
error_t stage_flip(alloc_t* a)
{
	error_t err = ERR_OK;
	float_t dist;
	float_t rl_avg, rh_avg;
	uint_t rl, rh, cl, ch;
	unsigned char done, good_flip;

	done = 0;
	while (!done) {
		good_flip = 0;
		rl = ROW_INVALID;
		rh = ROW_INVALID;
		/* iterate through rows below average */
		rl = sel_lo_row(a->avg, a->m, rl);
		while (rl != ROW_INVALID) {
			/* iterate through rows above average */
			rh = sel_hi_row(a->avg, a->m, rh);
			while (rh != ROW_INVALID) {
				/* iterate through columns below lba */
				for (cl = 0; cl <= a->lba; cl++) {
					/* iterate through columns above lba */
					for (ch = a->lba + 1; ch < a->n; ch++) {
						/* if the selected giving fields cannot give
						anything, continue */
						if (a->x[rl][cl] == 0 || a->x[rh][ch] == 0)
							continue;
						/* get current distance between rows */
						dist = ABS(a->avg[rl] - a->avg[rh]);
						/* flip */
						a->x[rl][cl] -= 1; a->x[rh][cl] += 1;
						a->x[rh][ch] -= 1; a->x[rl][ch] += 1;
						/* get new distance */
						rl_avg = row_average(a, rl);
						rh_avg = row_average(a, rh);
						if (dist <= ABS(rl_avg - rh_avg)) {
							/* if the distance hasn't improved, flipback */
							a->x[rl][cl] += 1; a->x[rh][cl] -= 1;
							a->x[rh][ch] += 1; a->x[rl][ch] -= 1;
							good_flip = 0;
						} else {
							/* if the distance improved, start over */
							a->avg[rl] = rl_avg;
							a->avg[rh] = rh_avg;
							good_flip = 1;
							break;
						}
					}
					/* if a good flip has been made anywhere in the iteration,
					start over. */
					if (good_flip)
						break;
				}
				if (good_flip)
					break;
				rh = sel_hi_row(a->avg, a->m, rh);
			}
			if (good_flip)
				break;
			rl = sel_lo_row(a->avg, a->m, rl);
		}
		if (!good_flip)
			done = 1;
	}

	return err;
}


/* main logic */
error_t alloc_run(alloc_t* a)
{
	error_t err = ERR_OK;
	
	err = stage_init(a);
	if (err == ERR_OK)
		err = stage_spread(a);
	if (err == ERR_OK)
		err = stage_flip(a);

	return 0;
}

/* returns the matrix */
uint_t** alloc_result_matrix(alloc_t* a)
{
	return a->x;
}

/* returns te averages */
float_t* alloc_result_avgs(alloc_t* a)
{
	return a->avg;
}

