/*
  ealp.h: equity allocation algorithm header.
  Copyright (c) 2000, Igor Zlatkovic. No rights to reserve.

  The problem is as follows:
	
 	    C1  C2  ... Cn    Cn+1
       +-----------------+------+-----
  R1   | X11 X12 ... X1n | Sr1  | Avg1
  R2   | X21 X22 ... X2n | Sr2  | Avg2
  ...  | ... ... ... ... | ...  | ...
  Rm   | Xm1 Xm2 ... Xmn | Srm  | Avgm
       +-----------------+------+-----
  Rm+1 | Sc1 Sc2 ... Scn | St   | At 
       +-----------------+------+-----
       | P1  P2  ... Pn  |

  You have a matrix with m rows and n columns, labeled R1-Rm 
  and C1-Cm respectively. Each column Ci has a constant Pi 
  associated with it, and each row Ri has an average value 
  Avgi associated with it. Values P1-Pi and Avg1-Avgi are 
  assumed to be real numbers (floating point values) and 
  matrix entries X11-Xmn are assumed to be natural numbers 
  (unsigned integers).

  Sri is the sum of the values in the row i, and Sci is the 
  sum of the values in the column i. Value St is therefore
  the sum of the row m+1 and the sum of the column n+1 at the 
  same time. The algorithm requires these two sums to be the 
  same, which is enforced by putting their values into the 
  same variable. Means, for any row i:
  
  Sri = Sum[Xij, {j,1,n}].
  Sci = Sum[Xji, {j,1,m}].

  and

  St = Sum[Scj, {j,1,m}].

  Average value Avgi of each row i is calculated as
  
  Avgi = Sum[Xij*Pj, {j,1,n}] / Sri. 
  
  Value At is the average value for the row Rm+1 and is 
  calulated as
  
  At = Sum[Scj*Pj, {j,1,n}] / St.

  Now, given p-values P1...Pn, sums of rows Sr1...Srm and
  sums of columns Sc1...Scn, find matrix X11-Xmn, such that
  average values Avg1-Avgm remain so close to each other
  as possible, ideally equal to At.

  The problem is believed to be NP-hard and an algorithm 
  which guarantees to find an optimal solution in acceptable 
  time is not known. The following algorithm is one possible 
  solution which finds the local minimum given a start value
  and how good te minimum is, depends on the start value. The
  algorithm will give good results, acceptable for most
  applications.

  Here is how it is used:

  You set up input values in some other way, of course, but
  the rest is about the same like this:

  uint_t sr[] = { 12, 14, 16, 15, 5, 8 };
  uint_t sc[] = { 6, 14, 22, 18, 10 };
  float_t p[] = { 51.0, 52.0, 53.0, 54.0, 55.0 };
  
  alloc_t* a = alloc_create(6, 5, p, sr, sc);
  if (a != NULL) {
      
      // run the algorithm
      alloc_run(a);

      // get the results
      uint_t** mtx = alloc_result_matrix(a);
      float_t* avg = alloc_result_avgs(a);

      alloc_destroy(a);
  }

  Have fun.
*/

#ifndef __EALP_H_
#define __EALP_H_

/* float_t is the floating point type to be used. */
typedef double float_t;

/* uint_t is the unsigned integer type to be used. */
typedef unsigned int uint_t;

/* local types */
typedef int error_t;

/* Allocation algorithm type, holding all information needed for
   the single run. */
struct _alloc_t;
typedef struct _alloc_t alloc_t;

/* Local constants */
#define ROW_INVALID ((uint_t)(-1))
#define COL_INVALID ((uint_t)(-1))

#define ERR_OK ((error_t)0)
#define ERR_WARN ((error_t)1)
#define ERR_FAIL ((error_t)(-1))

/* Function prototypes */
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

alloc_t* alloc_create(uint_t num_rows, uint_t num_cols,
	float_t* p, uint_t* row_sums, uint_t* col_sums);
void alloc_destroy(alloc_t* a);
error_t alloc_run(alloc_t* a);
uint_t** alloc_result_matrix(alloc_t* a);
float_t* alloc_result_avgs(alloc_t* a);
void alloc_display(alloc_t* a);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __EALP_H_ */

